--[[
Copyright 2010, 2011 William Matthews. All rights reserved.

See LICENSE for license terms.
]]

local CB = select(2, ...)
CB = LibStub("AceAddon-3.0"):NewAddon(CB, "CostBasis")
CB.Private = {}
local P = CB.Private
setmetatable(P,{__index=CB,__tostring=function() return "CostBasis" end})

for _,lib in pairs({"AceConsole-3.0","AceEvent-3.0","AceTimer-3.0"}) do
	LibStub(lib):Embed(P)
end

P.L = LibStub("AceLocale-3.0"):GetLocale("CostBasis",false)
local L = P.L

local tinsert, tsort = table.insert, table.sort

local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

local DS = DataStore

local dbDefaults = {
	global = {
		debug = false,
		tooltip = true,
		chatlog = true,
	},
	factionrealm = {
		characters = {
			['*'] = true
		},
		guilds = {
			['*'] = false
		},
		itemcount = {
			['*'] = 0,
		},
		itemcostbasis = {},
		itemadjustment = {},
	}
}

local optionsTable = {
	type = "group",
	args = {
		characters = {
			type = "multiselect",
			name = "Characters",
			desc = "Characters to include in cost basis calculations.",
			order = 0,
			values = function() 
				local rv = {}
				for characterName, character in pairs(DS:GetCharacters()) do
					if DS:GetCharacterFaction(character)==UnitFactionGroup("player") then
						rv[character]=DS:GetColoredCharacterName(character)
					end
				end
				return rv
			end,
			get = function(option,character) 
				return P.FR.characters[character] 
			end,
			set = function(option,character,state) 
				P:DebugPrint({character=character,state=state})
				P.FR.characters[character] = state
			end
		},
		guilds = {
			type = "multiselect",
			name = "Guild Banks",
			desc = "Guild banks to include in cost basis calculations.",
			order = 1,
			values = function()
				local rv = {}
				local factionGuilds = {}
				for characterName, character in pairs(DS:GetCharacters()) do
					if DS:GetCharacterFaction(character)==UnitFactionGroup("player") then
						local guild = DS:GetGuildInfo(character)
						if guild then
							factionGuilds[guild] = true
						end
					end
				end
				for guildName, guild in pairs(DS:GetGuilds()) do
					if factionGuilds[guildName] then
						rv[guild]=guildName
					end
				end
				return rv
			end,
			get = function(option,guild) 
				return P.FR.guilds[guild] 
			end,
			set = function(option,guild,state) 
				P:DebugPrint({guild=guild,state=state})
				P.FR.guilds[guild] = state
			end
		},
		tooltip = {
			type = "toggle",
			name = "Show Information in Tooltips",
			desc = "Show cost basis in item tooltips.",
			order = 100,
			get = function(option)
				return P.G.tooltip
			end,
			set = function(option,state)
				P.G.tooltip=state
			end
		},
		chatlog = {
			type = "toggle",
			name = "Display Transactions",
			desc = "Display transactions in chat frame.",
			order = 200,
			get = function(option)
				return P.G.chatlog
			end,
			set = function(option,state)
				P.G.chatlog=state
			end
		},
		debug = {
			type = "toggle",
			name = "Debug",
			desc = "Enable debugging messages.",
			order = 1000,
			get = function(option)
				return P.G.debug
			end,
			set = function(option,state)
				P.G.debug=state
			end
		},
	}
}

local SLASH_VERBS = {}
local SLASH_HELP = {}
local function slashHelp()
	for _,h in pairs(SLASH_HELP) do
		P:Print(h)
	end	
end
local function onSlashCB(arg,source)
	local pos=1
	local function nextArg()
		local rv
		rv, pos=P:GetArgs(arg,1,pos,arg)
		return rv
	end	
	local verb = nextArg()
	if verb and SLASH_VERBS[verb] then
		SLASH_VERBS[verb](nextArg)
	else
		slashHelp()
	end
end
function P:RegisterSlashVerb(verb,args,help,func)
	SLASH_VERBS[verb]=func
	SLASH_HELP[verb]=("/cb %s%s:  %s."):format(verb,args and " "..args or "",help)
end
P:RegisterSlashVerb("help",nil,"lists commands",slashHelp)


local PARSE_GSC={
	-- ###g ###s ###c
	{"^%s*(%d+%.?%d*)%s*g%s*(%d+%.?%d*)%s*s%s*(%d+%.?%d*)%s*c%s*$",function(g,s,c) return (tonumber(g)*100+tonumber(s))*100+tonumber(c) end},
	-- ###g ###s
	{"^%s*(%d+%.?%d*)%s*g%s*(%d+%.?%d*)%s*s%s*$",function(g,s) return (tonumber(g)*100+tonumber(s))*100 end},
	-- ###g ###c
	{"^%s*(%d+%.?%d*)%s*g%s*(%d+%.?%d*)%s*c%s*$",function(g,c) return tonumber(g)*10000+tonumber(c) end},
	-- ###g
	{"^%s*(%d+%.?%d*)%s*g%s*$",function(g) return tonumber(g)*10000 end},
	-- ###s ###c
	{"^%s*(%d+%.?%d*)%s*s%s*(%d+%.?%d*)%s*c%s*$",function(s,c) return tonumber(s)*100+tonumber(c) end},
	-- ###s
	{"^%s*(%d+%.?%d*)%s*s%s*$",function(s) return tonumber(s)*100 end},
	-- ###c
	{"^%s*(%d+%.?%d*)%s*c%s*$",function(c) return tonumber(c) end},
}
local function parseGSC(str)
	for _,p in ipairs(PARSE_GSC) do
		local s=p[1]
		local f=p[2]
		a,b,c=str:match(s)
		if a then return f(a,b,c) end
	end
end

local function slashSet(nextArg)
	local item=nextArg()
	local cost=nextArg()
	local t=nextArg()
	while t and t~="" do
		cost=cost..t
		t=nextArg()
	end
	local itemString=item
	local costString=cost
	cost=parseGSC(cost)
	item=P.idOfItem(item)
	if cost and item then
		P:SetDBItemCostBasis(item,cost)
		P:SetDBItemAdjustment(item,0)
		P:SetDBItemCount(item,P:GetDBItemCount(item) or 0)
	else
		P:DebugPrint({cost=cost,item=item,costString=costString,itemString=itemString})
		slashHelp()
	end
end

P:RegisterSlashVerb("set","item cost","sets the cost basis for an item",slashSet)

local initRelativeValues -- defined below

function CB:OnInitialize()
	-- DB init
	P.db = LibStub("AceDB-3.0"):New("CostBasisDB",dbDefaults)
	P.FR = P.db.factionrealm
	P.G = P.db.global
	
	-- Config init
	AceConfig:RegisterOptionsTable("CostBasis",optionsTable)
	local category=AceConfigDialog:AddToBlizOptions("CostBasis","Cost Basis")
	P:RegisterChatCommand("costbasis",onSlashCB)
	P:RegisterChatCommand("cb",onSlashCB)
	P:RegisterSlashVerb("config",nil,"opens configuration dialog",function()
		InterfaceOptionsFrame_OpenToCategory(category)
	end)
	
	initRelativeValues()
	
	P:Print("Loaded!")
end

-- x==x fails for nan!
local inf=1/0
local function isfinite(x)
	return x and type(x)=="number" and x==x and x~=inf and x~=-inf
end


function P:GetDBItemCount(id)
	return self.FR.itemcount[id]
end
function P:SetDBItemCount(id,count)
	self.FR.itemcount[id]=count
end
function P:GetDBItemCostBasis(id)
	return self.FR.itemcostbasis[id]
end
function P:SetDBItemCostBasis(id,cb)
	self.FR.itemcostbasis[id]=cb
end
function P:GetDBItemAdjustment(id)
	return self.FR.itemadjustment[id]
end
function P:SetDBItemAdjustment(id,adj)
	self.FR.itemadjustment[id]=adj
end

function CB:QueryItem(item,count)
	local id=P.idOfItem(item)
	if not id then return nil end
	count = count or 1
	local cb=P:GetDBItemCostBasis(id)
	return cb and cb*count
end

-- fromItems, toItems, toExpected: { itemId: number, ...}
-- "expected" defaults to "to", goldCost defaults to 0
function CB:RecordTransaction(goldCost,fromItems,toItems,toExpected)
	goldCost = goldCost or 0
	fromItems = fromItems or {}
	toItems = toItems or {}
	if toExpected then
		for k in pairs(toItems) do
			if toExpected[k]==nil then
				P:DebugPrint({unexpectedItem=P.linkOfItem(k),id=k})
				toExpected=toItems
				break
			end
		end
	else
		toExpected=toItems
	end
	assert(type(goldCost)=="number")
	P.Timer.new(function ()
		local cb = goldCost
		local unknown = {}
		for item, count in pairs(fromItems) do
			local icb = self:QueryItem(item,count)
			if not icb then
				table.insert(unknown,item)
			end
			cb = cb and icb and cb+icb
	--		self:RecordSellItem(fromItem.item,fromItem.count)
		end
		if not cb then 
			P:LogUnknown(unknown)
			return
		end
		local relValues = P.relativeValues(toExpected)
	
		--[[
		CB:DebugPrint({
			toItems = toItems,
			relValues=relValues,
		})
		]]
	
		for item in pairs(toExpected) do
			local itemCB = cb * relValues[item]
			assert(isfinite(itemCB),"itemCB="..tostring(itemCB).." cb="..tostring(cb).." relValues[item]="..tostring(relValues[item]))
			local count = toItems[item] or 0
			if count==0 then
				P:RecordAdjustItem(itemCB,item)
			else
				P:RecordBuyItem(itemCB,item,count)
			end
		end
		P:LogTransaction(goldCost,fromItems,toItems,toExpected)
	end,P.EPSILON_TIME)
end

-- must be called from a timer!
function P:RecordAdjustItem(cost,id)
	assert(type(cost)=="number","cost is not a number")
	assert(type(id)=="number","id is not a number")
	local count=math.min(self:GetDBItemCount(id) or 0,self:GetTotalCBItemCount(id) or 0)
	if not self:GetDBItemCostBasis(id) then
		count=0
	end
	self:SetDBItemCount(id,count)
	if count > 0 then
		self:SetDBItemCostBasis(id,self:GetDBItemCostBasis(id) + cost/count)
	else
		local adj=self:GetDBItemAdjustment(id)
		adj=(adj and adj+cost) or cost
		self:SetDBItemAdjustment(id,adj)
	end
end
-- must be called from a timer!
function P:RecordBuyItem(cost,id,count)
	assert(type(cost)=="number")
	assert(type(id)=="number")
	assert(type(count)=="number")
	assert(count>0)
	local owned = self:GetTotalCBItemCount(id)
	local prevCost = self:GetDBItemCostBasis(id)
	local prevCount = math.max(math.min(self:GetDBItemCount(id),owned-count),0)
	local adjustment = self:GetDBItemAdjustment(id) or 0
	local totalCost = (prevCost or 0)*prevCount + cost + adjustment
	local totalCount = prevCount + count
	assert(totalCount>0,'totalCount='..tostring(totalCount))
	local totalCB = totalCost / totalCount
	self:SetDBItemCostBasis(id,totalCB)
	self:SetDBItemCount(id,totalCount)
	self:SetDBItemAdjustment(id,nil)
end

function P.itemIconLink(item)
	local tex = ("|T%s:0|t"):format(GetItemIcon(item))
	local link = P.linkOfItem(item)
	return link:gsub("|h.*|h","|h"..tex.."|h")
end

-- items, expected: {id: count, ...}
function P.itemListToIconString(items,expected)
	items = items or {}
	expected = expected or items
	local rv=nil
	for item,exp in pairs(expected) do
		local tex = P.itemIconLink(item)
		local str
		local count = items[item] or 0
		if count~=exp then
			str=string.format("%sx%g[%g]",tex,count,exp)
		elseif count~=1 then
			str=string.format("%sx%g",tex,count)
		else
			str=tex
		end
		local cb=CB:QueryItem(item)
		if cb then
			str=string.format("%s (%s)",str,P.moneyString(cb))
		end
		rv = rv and rv..", "..str or str
	end
	return rv
end


local lastUnknown = nil
local unknownTimer = nil
local function clearUnknown()
	lastUnknown=nil
end
function P:LogUnknown(unknown)
	if not P.G.chatlog then return end
	local s = nil
	for _, item in ipairs(unknown) do
		local tex = P.itemIconLink(item)
		s = s and s..", "..tex or tex
	end
	if s~=lastUnknown then
		P:Print(format("Unknown cost basis for %s.  Cannot compute cost basis for result.",s))			
		lastUnknown=s
	end
	unknownTimer=P.Timer.updateOrNew(unknownTimer,clearUnknown,60)
end

function P:LogTransaction(goldCost,fromItems,toItems,toExpected)
	if not P.G.chatlog then return end
	goldCost=goldCost or 0
	assert(type(goldCost)=="number")
	local from = P.itemListToIconString(fromItems)
	local to = P.itemListToIconString(toItems,toExpected)
	local fromCost = goldCost>0 and P.moneyString(goldCost)
	local toCost = goldCost<0 and P.moneyString(-goldCost)
	from = (from and fromCost and fromCost..", "..from) or from or fromCost or ""
	to = (to and toCost and toCost..", "..to) or to or toCost or ""
	self:Print(string.format("%s |TInterface\\Glues\\Login\\UI-BackArrow.blp:0|t %s",to,from))
end

local valueMethodsByName = {}
local valueMethods = {{name="uniform",accuracy=0,method=function() return 1 end}}

local function valueMethodCompare(a,b)
	return a.accuracy>b.accuracy
end

-- accuracy between 1 and 100, 100 being most accurate
-- func(itemID) returns value or nil
function P.registerValueMethod(name,accuracy,func)
	if not name or type(name)~="string" then return nil, "invalid name" end
	if not accuracy or type(accuracy)~="number" or accuracy<1 or accuracy>100 then return nil, "invalid accuracy" end
	local t=valueMethodsByName[name]
	if not t then
		t={}
		valueMethodsByName[name]=t
		tinsert(valueMethods,t)
	end
	t.name=name
	t.accuracy=accuracy
	t.method=func
	tsort(valueMethods,valueMethodCompare)
	P:DebugPrint(valueMethods)
	return true
end

function P.relativeValues(expected)
	local first=next(expected)
	if not first then return {} end -- 0 items
	if not next(expected,first) then return {[first]=1} end -- 1 item
	for _, t in ipairs(valueMethods) do
		local method=t.method
		local cache = {}
		local totalVal = 0
		local val
		for item,exp in pairs(expected) do
			val = method(item)
			cache[item] = val
			totalVal = totalVal and val and isfinite(val) and val>0 and totalVal+val*exp 
		end
		if totalVal then
			assert(isfinite(totalVal) and totalVal>0)
			local rv = P.ItemSet.new()
			for item,exp in pairs(expected) do
				rv[item]=cache[item]*exp / totalVal
			end
			return rv
		end
	end
	error("SHOULD NEVER GET HERE!")
end

function initRelativeValues()
	if select(4, GetAddOnInfo("Auc-Advanced")) == 1 then
		P.registerValueMethod("AucAdvancedMarket",80,function(id)
			return id and AucAdvanced and AucAdvanced.API.GetMarketValue(P.linkOfItem(id))
		end)
	end
	
	if select(4, GetAddOnInfo("Enchantrix")) == 1 then
		P.registerValueMethod("EnchantrixStatic",30,function(id)
			return id and Enchantrix and Enchantrix.Constants.StaticPrices[id]
		end)
	end
	
	if select(4, GetAddOnInfo("TheUndermineJournal")) == 1 then
		P.registerValueMethod("TUJMarket",60,function(id) 
			return id and (TUJMarketInfo(itemID) or {}).market 
		end)
		P.registerValueMethod("TUJMean",50,function(id) 
			return id and (TUJMarketInfo(itemID) or {}).marketaverage
		end)
	end
	
	if select(4, GetAddOnInfo("TradeSkillMaster_AuctionDB")) == 1 then
		P.registerValueMethod("TSMAuctionDBMarket",70,function(id) 
			return TSMAPI and TSMAPI:GetData("market", itemID) 
		end)
	end
	
	if select(4, GetAddOnInfo("Auctionator")) == 1 then
		P.registerValueMethod("AtrValue",65,function(id) 
			return Atr_GetAuctionBuyout(P.linkOfItem(id))
		end)
	end	
	
	-- TODO: add wowecon
	
	P.registerValueMethod("Vendor",10,function(id)
		return select(11, GetItemInfo(id))
	end)
end
